#################
## EXAMPLE: simple Coordinate class
#################
class Coordinate(object):
    """ A coordinate made up of an x and y value """
    def __init__(self, x, y):
        """ Sets the x and y values """
        self.x = x
        self.y = y
    def __add__(self, other):
        return Coordinate(self.x+other.x, self.y+other.y)
    def __sub__(self, other):
        return Coordinate(self.x-other.x, self.y-other.y)
    def __str__(self):
        """ Returns a string representation of self """
        return "<" + str(self.x) + "," + str(self.y) + ">"
    def distance(self, other):
        """ Returns the euclidean distance between two points """
        x_diff_sq = (self.x-other.x)**2
        y_diff_sq = (self.y-other.y)**2
        return (x_diff_sq + y_diff_sq)**0.5


c = Coordinate(3,4)
origin = Coordinate(0,0)
#print(c.x, origin.x)
#print(c.distance(origin))
#print(Coordinate.distance(c, origin))
#print(origin.distance(c))
##print(c)
#c+origin


#################
## EXAMPLE: simple Circle class
#################

class Circle(object):
    def __init__(self, center, radius):
        self.center = center
        self.radius = radius
    def __str__(self):
        return "circle: "+str(self.center)+", "+str(self.radius)
    def is_inside(self, point):
        return point.distance(self.center)<self.radius
    
c = Circle(Coordinate(2, 2), 2)
p = Coordinate(1,1)
print(c.is_inside(p))

#################
## EXAMPLE: default parameters
#################

def f(x, y=1):
    return x+y

print(f(1, 4))
print(f(2))

#################
## EXAMPLE: BAD default parameters
## from https://docs.python-guide.org/writing/gotchas/
#################

def append_to(element, to=[]):
    to.append(element)
    return to

my_list = append_to(12)
print(my_list)

my_other_list = append_to(42)
print(my_other_list)

## you would expect teh second call to return [42] but that's not what you get

#################
## EXAMPLE: Credit Cards
#################



class CreditCard(object):
    def __init__(self, yearly_rate_in_percent):
        self.balance = 0
        self.yearly_rate_in_percent = yearly_rate_in_percent
        
    def get_balance(self):
        return self.balance
    
    def spend(self, amount):
        self.balance += amount
        
    def pay_bill(self, amount):
        self.balance -= amount
        
    def monthly_update(self):
        interests = self.get_balance()*self.yearly_rate_in_percent/12.0/100.0
        self.balance += interests
        
    def __str__(self):
        return "amount due: "+str(self.get_balance())

#alternative that stores the monthly multiplier
# follows the same specifications 
#  example of abstraction where internal representation 
# can be changed without affecting external behavior      
        
class CreditCard(object):
    def __init__(self, yearly_rate_in_percent):
        self.balance = 0
        self.monthly_interests_fraction = yearly_rate_in_percent/12.0/100.0
        
    def get_balance(self):
        return self.balance
    
    def spend(self, amount):
        self.balance += amount
        
    def pay_bill(self, amount):
        self.balance -= amount
        
    def monthly_update(self):
        interests = self.get_balance()*self.monthly_interests_fraction
        self.balance += interests
        
    def __str__(self):
        return "amount due: "+str(self.get_balance())

#alternative that stores a list of operations
# follows the same specifications    
class CreditCard(object):
    def __init__(self, yearly_rate_in_percent):
        self.operations = []
        self.monthly_interests_fraction = yearly_rate_in_percent/12.0/100.0
        
    def get_balance(self):
        return sum(self.operations)
    
    def spend(self, amount):
        self.operations.append(amount)
        
    def pay_bill(self, amount):
        self.operations.append(-amount)
        
    def monthly_update(self):
        interests = self.get_balance()*self.monthly_interests_fraction
        self.operations.append(interests)
        
    def __str__(self):
        return "amount due: "+str(self.get_balance())

my_card = CreditCard(20)
my_card.spend(1.99)
my_card.spend(3.99)
my_card.monthly_update()
my_card.pay_bill(4)
print(my_card)


#### REWARD CARD #########

class RewardCard(CreditCard):
    def __init__(self, yearly_rate_in_percent, miles_per_dollar):
        CreditCard.__init__(self, yearly_rate_in_percent)
        self.miles_per_dollar = miles_per_dollar
        self.miles = 0
        
    def get_miles(self):
        return self.miles
    
    def spend(self, amount):
        CreditCard.spend(self, amount)
        self.miles += amount * self.miles_per_dollar
        
    def __str__(self):
        return CreditCard.__str__(self)+"; reward "+str(self.get_miles())+" miles"


my_card = RewardCard(20, 0.01)
my_card.spend(1.99)
my_card.spend(3.99)
my_card.monthly_update()
my_card.pay_bill(4)
print(my_card)

#################
## EXAMPLE: Class variables
## from https://docs.python-guide.org/writing/gotchas/
#################


#################################
## Animal abstract data type 
#################################
class Animal(object):
    #class variables will be shared by all instances
    species = ""
    counter = 0
            #class variable will be shared by all instances
    def __init__(self, name, age):
        self.age = age
        self.name = name
        #increment counter to count all amimals
        Animal.counter += 1
    def __str__(self):
        return self.species+" "+str(self.name)+":"+str(self.age)

    def get_age(self):
        return self.age
    def get_name(self):
        return self.name
    def set_age(self, newage):
        self.age = newage
    def set_name(self, newname=""):
        self.name = newname
    def get_total_number_of_animals(self):
        return Animal.counter
    #alternative: class function without self
    def class_get_total_number_of_animals():
        return Animal.counter
    # note how counter can be gotten but not set

class Rabbit(Animal):
    species = "Rabbit"
    #insert here whatever it is rabbits do
    
class Fox(Animal):
    species = "Fox"
    #insert here whatever it is foxes do
    
r = Rabbit('knufflebunny', 2)    
f1 = Fox("Goupil", 4)
f2 = Fox("Renard", 5)
print(r)
print(f1)
print(f2)
print(r.get_total_number_of_animals())
print(Animal.class_get_total_number_of_animals())

